
# This file is automatically generated during the generation of setup.py
# Copyright 2023-2026, Open XAI PnP
author = "Team SAIL"
author_email = "open.xai@gmail.com"
version = "0.1.3"
full_version = "0.1.3"
git_version = '78e93c7eecf360e21ffaf95a930d8790633be820'
release = True
homepage = "https://github.com/OpenXAIProject/pnpxai"
